from ase.io import read
import ase
import numpy as np
from ncnice import *
from rascal.representations import SphericalExpansion, SphericalInvariants
from rascal.utils import (
    WignerDReal,
    ClebschGordanReal,
    spherical_expansion_reshape,
    lm_slice,
)
from copy import deepcopy
import tqdm
from rascal.utils import (
    get_radial_basis_covariance,
    get_radial_basis_pca,
    get_radial_basis_projections,
    get_optimal_radial_basis_hypers,
)
from sklearn.linear_model import Ridge, RidgeCV
from feat_settings import *


def normalize(feats):
    feats_v = np.asarray(feats)
    num_frames = feats_v.shape[0]
    num_atoms = feats_v.shape[1]
    feats_v = feats_v.reshape(num_frames * num_atoms, -1)
    norm = np.sqrt((feats_v**2).sum(axis=1).mean())
    feats_v = feats_v.reshape(num_frames, num_atoms, -1)
    feats_v /= norm
    return feats_v


def atom_2_structure(x):
    """contracts aatomic descriptors into structure descriptors"""
    return np.concatenate([x[0], x[1], x[2:].sum(axis=0)], axis=0).reshape(-1)


frames = read(frames_file, ":")
for f in frames:
    # use same specie for all atoms so we get a single projection matrix,
    # which we can apply throughout. a bit less efficient but much more practical
    f.numbers = f.numbers * 0 + 11
spex_hypers = get_optimal_radial_basis_hypers(
    spex_hypers, frames, expanded_max_radial=20
)
# ... and we replicate it
pm = spex_hypers["optimization"]["RadialDimReduction"]["projection_matrices"][11]
spex_hypers["optimization"]["RadialDimReduction"]["projection_matrices"] = {
    i: pm for i in range(64)
}

spex_hypers["global_species"] = [
    1,
    11,
    17,
]  # we added 1 here because we add dummy particle labelled by 'H'

spex = SphericalExpansion(**spex_hypers)
hypers_ij = deepcopy(spex_hypers)
hypers_ij["expansion_by_species_method"] = "structure wise"
spex_gij = SphericalExpansion(**hypers_ij)

mycg = ClebschGordanReal(spex_hypers["max_angular"])
frames = read(frames_file, ":")
# all atoms but first two become inert H . extends the dataset by creating multiple substitutions
energy = []
distance = []
new_frames = []
for f in frames:
    f.symbols[
        2:
    ] = "H"  # dummy particle - can be assigned any symbol other than Na or Cl
    f.positions += 50 - f.cell[0, 0] / 2
    f.cell = [100, 100, 100]
for f in frames:
    for k in range(5):
        f.positions[0], f.positions[1], f.positions[2 * k], f.positions[2 * k + 1] = (
            f.positions[2 * k].copy(),
            f.positions[2 * k + 1].copy(),
            f.positions[0].copy(),
            f.positions[1].copy(),
        )
        f.info["dist"] = np.linalg.norm(f.positions[0] - f.positions[1])
        f.info["energy"] = 1.0 / f.info["dist"]
        new_frames.append(f.copy())
        energy.append(f.info["energy"])
        distance.append(f.info["dist"])
energy = np.asarray(energy)
distance = np.asarray(distance)
frames = new_frames

feats = {
    "rho_i^1;00": [],
    "rho_i^2;00": [],
    "rho_i^3;00": [],
    "rho_i^(1<-1);00": [],
}

for f in tqdm(frames[:]):
    rhoi = compute_rhoi(f, spex, spex_hypers)
    gij = compute_gij(f, spex_gij, hypers_ij)
    rho1i_l, prho1i_l = compute_rho1i_lambda(rhoi, 0, mycg)
    feats["rho_i^1;00"].append(
        atom_2_structure(rho1i_l[..., np.where(prho1i_l == 1)[0], :])
    )
    rho1ij = compute_rho1ij(rhoi, gij, mycg)
    rho2i_l, prho2i_l = compute_rho2i_lambda(rhoi, 0, mycg)
    feats["rho_i^2;00"].append(
        atom_2_structure(rho2i_l[..., np.where(prho2i_l == 1)[0], :])
    )
    rho2i_l_all, prho2i_l_all = compute_all_rho2i_lambda(rhoi, mycg)
    rho3i_l, prho3i_l = compute_rho3i_lambda(rho2i_l_all, rhoi, 0, mycg, prho2i_l_all)
    feats["rho_i^3;00"].append(
        atom_2_structure(rho3i_l[..., np.where(prho3i_l == 1)[0], :])
    )
    rho1ijp_l_all, prho1ijp_l_all = compute_all_rho1ijp_lambda(rhoi, gij, mycg)
    rho11ijp, prho11ijp = compute_rho11ijp_lambda(
        rhoi, rho1ijp_l_all, 0, mycg, prho1ijp_l_all
    )
    rho11P, prho11P = contract_rhoij(rho11ijp, prho11ijp, f.symbols, ["H", "Na", "Cl"])
    feats["rho_i^(1<-1);00"].append(
        atom_2_structure(rho11P[..., np.where(prho11P == 1)[0], :])
    )

for k in feats:
    feats[k] = normalize(feats[k])

nmax = spex_hypers["max_radial"]
y = []
for f in frames[: len(feats["rho_i^1;00"])]:
    y.append(f.info["energy"])
y = np.asarray(y)
nacl_results = {"ref": y}

mode = "nacl"  # "rho_i^1;00"

if mode == "nacl":
    x = np.hstack(
        [
            feats["rho_i^1;00"].reshape(len(y), -1)[:, 1 * nmax : 3 * nmax],
            feats["rho_i^1;00"].reshape(len(y), -1)[:, 4 * nmax : 6 * nmax],
        ]
    )
else:
    x = feats[mode].reshape(len(y), -1)

rcv = RidgeCV(alphas=np.geomspace(1e-5, 1e4, 40), cv=4)
rcv.fit(x[::2], y[::2])
try:
    print("alpha ", rcv.alpha_)
except:
    pass
yp = rcv.predict(x)
nacl_results[mode] = yp
print(np.sqrt(((y[1::2] - yp[1::2]) ** 2).mean()) / np.std(y[1::2]))
np.save("nacl-dummy_results.npy", nacl_results)
