from ase.io import read, write
import ase
import json
import hickle
import pickle
from tqdm import tqdm


class tqdm_reusable:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def __iter__(self):
        return tqdm(*self._args, **self._kwargs).__iter__()


from copy import deepcopy
import numpy as np
import scipy as sp

from rascal.representations import SphericalExpansion, SphericalInvariants
from rascal.utils import (
    get_radial_basis_covariance,
    get_radial_basis_pca,
    get_radial_basis_projections,
    get_optimal_radial_basis_hypers,
)
from rascal.utils import radial_basis
from rascal.utils import (
    WignerDReal,
    ClebschGordanReal,
    spherical_expansion_reshape,
    lm_slice,
)

from ncnice import *
import sys, os

sys.path.append(os.getcwd())
from feat_settings import *


"""
imports:
spex_hypers (spherical expansion parameters)
lambda_max (maximum order of lambda-features)

frames_file (name of frames)
"""

print("Starting up")
spex = SphericalExpansion(**spex_hypers)
mycg = ClebschGordanReal(spex_hypers["max_angular"])

if feat_slice is None:
    feat_slice = slice(None)

print("Computing optimal radial basis")
frames = read(frames_file, feat_slice)
for f in frames:
    f.numbers = f.numbers * 0 + 11

spex_hypers = get_optimal_radial_basis_hypers(
    spex_hypers, frames[::pca_stride], expanded_max_radial=16
)
# ... and we replicate it
pm = spex_hypers["optimization"]["RadialDimReduction"]["projection_matrices"][11]
spex_hypers["optimization"]["RadialDimReduction"]["projection_matrices"] = {
    i: pm for i in range(64)
}
spex = SphericalExpansion(**spex_hypers)

# these are to compute |r_ij; g> terms
hypers_ij = deepcopy(spex_hypers)
hypers_ij["expansion_by_species_method"] = "structure wise"
spex_gij = SphericalExpansion(**hypers_ij)

frames = read(frames_file, feat_slice)
feats = {
    "rho_i^1;00": [],
    "rho_i^2;00": [],
    "rho_i^3;00": [],
    "rho_i^(1<-1);00": [],
}


def compress_naclfeat(feat):
    """compress frame-wise NaCl feats"""
    feat = np.asarray(feat).reshape((64, -1))
    # since the dataframe has in total 64 Na-Cl atoms in alternating order
    return np.concatenate([feat[::2], feat[1::2]], axis=1).sum(axis=0)


print("Computing features")
for f in tqdm(frames[:]):
    rhoi = compute_rhoi(f, spex, spex_hypers)
    gij = compute_gij(f, spex_gij, hypers_ij)
    rho1i_l, prho1i_l = compute_rho1i_lambda(rhoi, 0, mycg)
    feats["rho_i^1;00"].append(
        compress_naclfeat(rho1i_l[..., np.where(prho1i_l == 1)[0], :])
    )

    rho1ij = compute_rho1ij(rhoi, gij, mycg)
    rho2i_l, prho2i_l = compute_rho2i_lambda(rhoi, 0, mycg)
    feats["rho_i^2;00"].append(
        compress_naclfeat(rho2i_l[..., np.where(prho2i_l == 1)[0], :])
    )

    rho2i_l_all, prho2i_l_all = compute_all_rho2i_lambda(rhoi, mycg)
    rho3i_l, prho3i_l = compute_rho3i_lambda(rho2i_l_all, rhoi, 0, mycg, prho2i_l_all)
    feats["rho_i^3;00"].append(
        compress_naclfeat(rho3i_l[..., np.where(prho3i_l == 1)[0], :])
    )

    rho1ijp_l_all, prho1ijp_l_all = compute_all_rho1ijp_lambda(rhoi, gij, mycg)
    rho11ijp, prho11ijp = compute_rho11ijp_lambda(
        rhoi, rho1ijp_l_all, 0, mycg, prho1ijp_l_all
    )
    rho11P, prho11P = contract_rhoij(rho11ijp, prho11ijp, f.symbols, ["Na", "Cl"])
    feats["rho_i^(1<-1);00"].append(
        compress_naclfeat(rho11P[..., np.where(prho11P == 1)[0], :])
    )

pickle.dump(feats["rho_i^1;00"], open(feat_prefix + "rho1i-feat.pickle", "wb"))
pickle.dump(feats["rho_i^2;00"], open(feat_prefix + "rho2i-feat.pickle", "wb"))
pickle.dump(feats["rho_i^3;00"], open(feat_prefix + "rho3i-feat.pickle", "wb"))
pickle.dump(feats["rho_i^(1<-1);00"], open(feat_prefix + "rho11MPi-feat.pickle", "wb"))
