import numpy as np
from ase.io import read, write
import ase
import scipy as sp

from rascal.representations import SphericalExpansion, SphericalInvariants
from rascal.utils import (
    get_radial_basis_covariance,
    get_radial_basis_pca,
    get_radial_basis_projections,
    get_optimal_radial_basis_hypers,
)
from rascal.utils import radial_basis
from rascal.utils import (
    WignerDReal,
    ClebschGordanReal,
    spherical_expansion_reshape,
    lm_slice,
)

from ncnice import *
import sys, os

sys.path.append(os.getcwd())


def get_error(first, second):
    return np.sqrt(np.mean((first - second) ** 2))


def get_spread(values):
    return np.sqrt(np.mean((values - np.mean(values)) ** 2))


def get_kernel(feat0, zeta):
    # assume feat0 shape is (nframes, featshape)i.e. The inpute here has been summed over the atom index.
    return (feat0 @ feat0.T) ** zeta


def test_regress_cv(
    feat,
    y,
    itrain,
    itest,
    alphas=[1e-12],
    tf=0.5,
    cv=None,
    fit_intercept=True,
    around_mean=False,
    sample_weight=None,
):
    feat.shape = (feat.shape[0], -1)
    print("feat shape", feat.shape)
    ntrain = int(len(feat) * tf)
    itest.sort()
    itrain.sort()
    clf = RidgeCV(alphas=alphas, cv=cv, fit_intercept=fit_intercept)
    if around_mean == False:
        print("NOT predicting around_mean")
        clf.fit(feat[itrain], y[itrain], sample_weight)
        print("Best regularizer ", clf.alpha_, " ntrain", ntrain, file=sys.stderr)
        predictions = clf.predict(feat)
        print(
            "Train error: %10.5f;   Test error: %10.5f"
            % (
                get_error(predictions[itrain], y[itrain]) / get_spread(y),
                get_error(predictions[itest], y[itest]) / get_spread(y),
            ),
            file=sys.stderr,
        )
        return predictions[itrain], predictions[itest]
    else:
        print("predicting around_mean")
        mean_train = np.mean(y[itrain])
        clf.fit(feat[itrain], y[itrain] - mean_train, sample_weight)
        print("Best regularizer ", clf.alpha_, " ntrain", ntrain, file=sys.stderr)
        predictions = clf.predict(feat) + mean_train
        print(
            "Train error: %10.5f;   Test error: %10.5f"
            % (
                get_error(predictions[itrain], y[itrain]) / get_spread(y),
                get_error(predictions[itest], y[itest]) / get_spread(y),
            ),
            file=sys.stderr,
        )
        return predictions[itrain], predictions[itest]


def test_regress(
    feat, y, itrain, itest, alpha=1e-12, tf=0.5, fit_intercept=True, around_mean=False
):
    feat.shape = (feat.shape[0], -1)
    print("feat shape", feat.shape)
    ntrain = int(len(feat) * tf)
    itest.sort()
    itrain.sort()
    clf = Ridge(alpha=alpha, fit_intercept=fit_intercept)
    if around_mean == False:
        print("NOT predicting around_mean")
        clf.fit(feat[itrain], y[itrain])
        print(" ntrain", ntrain, file=sys.stderr)
        predictions = clf.predict(feat)
        print(
            "Train error: %10.5f;   Test error: %10.5f"
            % (
                get_error(predictions[itrain], y[itrain]) / get_spread(y),
                get_error(predictions[itest], y[itest]) / get_spread(y),
            ),
            file=sys.stderr,
        )
        return predictions[itrain], predictions[itest]
    else:
        print("predicting around_mean")
        mean_train = np.mean(y[itrain])
        clf.fit(feat[itrain], y[itrain] - mean_train)
        print(" ntrain", ntrain, file=sys.stderr)
        predictions = clf.predict(feat) + mean_train
        print(
            "Train error: %10.5f;   Test error: %10.5f"
            % (
                get_error(predictions[itrain], y[itrain]) / get_spread(y),
                get_error(predictions[itest], y[itest]) / get_spread(y),
            ),
            file=sys.stderr,
        )
        return predictions[itrain], predictions[itest]


def test_regress_kernel(kernel, y, itrain, itest, alphas=[1e-12], tf=0.5, cv=3):
    ntrain = len(itrain)
    krcv = KernelRidge(kernel="precomputed")
    gscv = GridSearchCV(krcv, dict(alpha=alphas), cv=cv)  # ,scoring=scoring)
    ktrain = kernel[itrain][:, itrain]
    # ktest = kernel[itrain][:, itest]
    gscv.fit(ktrain, y[itrain])
    cv_stats = gscv.cv_results_
    alpha = gscv.best_params_["alpha"]
    clf = KernelRidge(kernel="precomputed", alpha=alpha)
    clf.fit(ktrain, y[itrain])
    print("Best regularizer ", clf.alpha, " ntrain", ntrain, file=sys.stderr)
    predictions = clf.predict(kernel[:][:, itrain])
    print("predictions.shape", predictions.shape)
    # predict (x) where X has shape [n_samples, n_samples_fitted], where n_samples_fitted is the number of samples used in the fitting for this estimator.
    print(
        "Train error: %10.5f;   Test error: %10.5f"
        % (
            get_error(predictions[itrain], y[itrain]) / get_spread(y),
            get_error(predictions[itest], y[itest]) / get_spread(y),
        ),
        file=sys.stderr,
    )
    return predictions, predictions[itest]
